// Simulation des chaines de Markov

function [x] = tirer(proba)
   // renvoie un nombre tir au hasard selon la proba
   r = rand(1,1,'uniform');
   res = find([cumsum(proba)] > r);
   x = res(1,1)
endfunction;
// tirer(P(1,:))

function [traj] = trajectoire(P,depart,nb_pas)
   // trajectoire d'une chaine de Markov
   traj = zeros(nb_pas,1)
   traj(1,1) = depart
   for i=2:nb_pas,
      traj(i,1) = tirer(P(traj(i-1,1),:))
   end;
endfunction;
// plot2d(trajectoire(P,2,100))

// Matrice potentielle
// On remarque si x -> y, cela est fait par un chemin de longueur < nb tats
// par consquent, il suffit de calculer une somme borne
// La matrice potentielle fournit les connections dans le graphe 

function [u] = potentielle(P,k)
   // matrice potentielle
   [n,m] = size(P);
   u = eye(n,m);
   temp = u;
   for i = 1:k,
      temp = temp * P;
      u = u + temp;
   end;
endfunction;
// potentielle(P,10)

// une chaine est irrductible 
// ssi les termes de u sont tous > 0
function [x] = irreductible(P)
   [n,m] = size(P);
   pk = eye(n,m);
   u = pk;
   p = (P>0)*1;
   for i = 1:n,
      pk = ((pk * p)>0)*1;
      u = u + pk;
   end;
   x = prod((u > 0)*1);
endfunction;
// irreductible(P)


// Si une matrice est irrductible (rcurrente positive dans le cas fini)
// Intgrale le long d'une trajectoire
// on suppose f donne par ses chantillons par le vecteur des f(i)

// Fonction indicatrice
function [f] = indicatrice(P,i)
   // f est l'indicatrice de l'tat i : 1_{i}
   [n,m] = size(P);
   f = zeros(n,1);
   f(i,1) = 1;
endfunction;
// f = indicatrice(P,2)

function [x] = integrale(f,P,n)
   // retourne le vecteur des intgrales partielles : 1/p * \sum_{k=1}^p f(X_k)
   traj = trajectoire(P,1,n);
   num = cumsum(f(traj));
   den = 1 ./ cumsum(ones(n,1));
   x = num .* den;
endfunction;
// plot2d(integrale(f,P,500))

// Calcul de la proba invariante  l'aide de la matrice potentielle
function [mu] = proba(P,k)
   [n,m] = size(P);
   u = potentielle(P,k-1);
   mu = sum(u,'r') / (k * n)
endfunction
// proba(P,100)

// Ou thorme de Perron-Froebenius
function [mu] = probap(P,k)
   [n,m] = size(P);
   mu = sum(P^k,'r') / n
endfunction;
// probap(P3,100)

// Ou encore calcul exact par rsolution de systme linaire
function [mu] = probas(P)
   [n,m] = size(P);
   A = P' - eye(m,n);
   A(1,:) = ones(1,n);
   mu = (A\( [1;zeros(n-1,1)] ))';
endfunction;
// probas(P3)


// Comparaison des vitesses de convergence
function [] = plot_convergence(P,fin)
   [n,m] = size(P);
   temp = zeros(fin,2*n);
   u = eye(n,m);
   pn = u;
   for i=1:fin,
      pn = pn * P;
      u = u + pn;
      temp(i,:) = [(sum(u,'r') / (i * n)) (sum(pn,'r')/n)];
   end;
   msg = "LGN(" + strcat(string(1:n),")@LGN(") + ")@PF(" + strcat(string(1:n),")@PF(") + ")";
   plot2d(1:fin,temp,leg=msg);
   xtitle(["Convergence vers la mesure invariante" "Comparaison entre la Loi des Grands Nombres et le Thorme de Perron-Froebenius"],"trajectoire","");
endfunction;
// plot_convergence(P,50)

// Recherche des priodes
// la priode est le pgcd des longueurs des circuits
// d divise tout cycle <=> d divise tout circuit
function [g] = periode(P)
   [n,m] = size(P);
   toutes_l = zeros(1,n); // initialis  0 car pgcd divise 0
   p = (P>0)*1;
   pk = eye(n,m);
   for k=1:n
      pk = ((pk * p)>0)*1;
      if (sum(eye(pk).*pk)>0) then
	 toutes_l(1,k) = k;
      end;
   end
   g = gcd(uint32(toutes_l));
endfunction

// Quelques matrices pour tester
P0 = eye(3,3) // matrice identit
P1 = [0 1 0; 0 0 1; 1 0 0] // matrice circulante
P2 = [0.9 0.1 0; 0 0.9 0.1; 0 0 1] // matrice triangulaire suprieure
P3 = [0.4  0.6  0.0;0.2  0.5  0.3;0.1  0.7  0.2;]
	// mto   1 neige   2 nuages   3 soleil
P4 = [0.5 0.5 0; 0 0.5 0.5; 0.5 0 0.5]
P5 = [0 1 0 0; 0 0 1 0; 1 0 0 1; 0 1 0 0];


// Utilisation des fonctions prcdentes

tirer(ones(1,6)/6) // lancer un d  6 faces
plot2d(trajectoire(P1,1,20)); // trajectoire d'une chaine de markov
plot2d(trajectoire(P2,1,20)); // 
plot2d(trajectoire(P3,1,40)); // 
f = indicatrice(P1,2) // fonction indicatrice de l'tat 2
plot2d(integrale(f,P1,50));    // calcul d'une intgrale
plot2d(integrale(f,P2,50));    
plot2d(integrale(f,P3,50));    
probap(P1,20) // calcul de proba invariante
probap(P2,20)
proba(P3,20)
probap(P3,20)
plot_convergence(P1,50);
plot_convergence(P2,50);
plot_convergence(P3,50);
periode(P1)
periode(P2)
periode(P3)


// Marche alatoires
// une proba donne un saut relatif

// Exemples de probas
P0 = [0.5 0 0.5] // marche alatoire symtrique
P1 = [0.55 0 0.45] 
P2 = [0.45 0 0.55] 

function [x] = tirer(proba)
   // renvoie un nombre tir au hasard selon la proba
   [n,m] = size(proba);
   r = rand(1,1,'uniform');
   res = find([cumsum(proba)] > r);
   x = res(1)-floor(m/2)-1
endfunction;
tirer(P1(1,:))

function [traj] = trajectoire(P,depart,nb_pas)
   // trajectoire d'une marche alatoire
   traj = zeros(nb_pas,1)
   traj(1,1) = depart
   for i=2:nb_pas,
      traj(i,1) = traj(i-1,1) + tirer(P(1,:))
   end;
endfunction;

tr = zeros(1000,100);
for i=1:100,
   tr(:,i) = trajectoire(P,0,1000);
end;
plot2d(tr(:,1)) // trac d'une trajectoire
plot2d(sum(sqrt(tr.*tr),'c')/100)

function [x,y] = distribution(v,a,b)
   // a et b sont des entiers
   x = (a:b)';
   y = zeros(x);
   for i=1:(b-a+1),
     y(i,1) = length(find(v==x(i,1)));
   end
   n=length(find((v>=a) *1 .* (v<=b)));
   y = y/n;
endfunction;
function [e] = esperance(v,a,b)
   // a et b sont des entiers
   [x,y] = distribution(v,a,b);
   e = sum(x .* y);
endfunction;

function [] = plot_convergence(P,depart,pas)
   // convergence en loi d'une marche alatoire
   // affiche les lois successives (restreintes  [a b])
   // values aux instants donns par le vecteur pas
   l = length(pas);
   tr = trajectoire(P, depart, pas(l));
   a = min(tr);
   b = max(tr);
   x = (a:b)';
   y = zeros(b-a+1,l);
   for i=1:l,
      [u,v] = distribution(tr(1:pas(i),1),a,b);
      y(:,i) = v;
   end
   msg = "n = " + strcat(string(pas),"@n = ");
   plot2d(x,y,leg=msg);
   xtitle("Convergence en loi de la distribution empirique")
endfunction;
plot_convergence(P0,0,[10 50 100 1000 10000])


// Systme ALOHA
// On recommence car la marche n'est pas uniforme en espace

A = [0.6 0.1 0.2 0.1 ] // probabilit d'avoir n nouveaux messages
A = [0.6 0.39 0.01] 

function [proba] = proba_aloha(i,a,nu)
   // renvoie l'tat suivant  partir de i
   [n,m] = size(a);
   if i == 0 then
      proba = [ 0, a(1) + a(2), 0 , a(3:$) ];
   else
      b = binomial(nu,i);
      proba = [ b(2)*a(1), (1-b(2))*a(1) + b(1)*a(2), (1-b(1))*a(2), a(3:$) ];
   end
endfunction;
proba_aloha(0,A,0.1)

function [] = dessine_proba_aloha(n,a,nu)
   y_dessin = zeros(n+1,length(a)+1)
   for i=1:(n+1),
      y_dessin(i,:) = proba_aloha(i-1,a,nu);
   end
   msg = "offset = " + strcat(string(-1:length(a)-1),"@offset = ");
   plot2d(0:n,y_dessin,leg=msg)
endfunction
dessine_proba_aloha(20,A,0.1)

function [traj] = trajectoire(anu,depart,nb_pas)
   // trajectoire d'un ALOHA instable
   a = anu(1:($-1));
   nu = anu($);
   traj = zeros(nb_pas,1)
   traj(1,1) = depart
   for i=2:nb_pas,
      traj(i,1) = traj(i-1,1) + tirer(proba_aloha(traj(i-1,1), a, nu))-2
   end;
endfunction;
plot2d(trajectoire([A 0.5],0,1000))
plot_convergence([A 0.5],0,[10 50 100 10000])

function [traj] = trajectoire(anu,depart,nb_pas)
   // trajectoire d'un ALOHA stable
   a = anu;
   traj = zeros(nb_pas,1)
   traj(1,1) = depart
   for i=2:nb_pas,
      nu = (a(1)-a(2)) / (traj(i-1,1)*a(1) - a(2));
      traj(i,1) = traj(i-1,1) + tirer(proba_aloha(traj(i-1,1), a, nu))-2
   end;
endfunction;
plot2d(trajectoire(A,0,1000))
plot_convergence(A,0,[10 50 100 1000 100000])

function [b] = stable(A)
   b = A * (0:(length(A)-1))' < A(1) * exp(A(2)/A(1)-1)
endfunction;

tr = zeros(1000,100);
for i=1:100,
   tr(:,i) = trajectoire(P,0,1000);
end;
plot2d(tr(:,1)) // trac d'une trajectoire
plot2d(sum(sqrt(tr.*tr),'c')/100)

// Files d'attente

function [traj] = trajectoire(anu,depart,nb_pas)
   // Trajectoire d'une file d'attente
   lambda = anu(1);
   p = anu(2);
   traj = zeros(nb_pas,1)
   traj(1,1) = depart
   for i=2:nb_pas,
      traj(i,1) = max(traj(i-1,1) - grand(1,1,'geom',p),0) + grand(1,1,'poi',lambda)
   end;
endfunction;

function [b] = stable(anu)
   lambda = anu(1);
   p = anu(2);
   b = lambda < (1-p)/p
endfunction;

x = 0.1:0.1:1;
y = (1 - x) ./ x;
plot2d(x,y);

A = [2 0.5]; // instable
A = [2 0.3]; // stable
stable(A)
plot2d(trajectoire(A,10,5000));
tr = trajectoire(A,10,5000);
esperance(tr,min(tr),max(tr))
plot_convergence(A,0,[10 50 100 1000 100000])


